# coding: utf-8
'''
Tämä on satumetsän peikkojen ohjelmoima simulaattori, jolla voi harjoitella
peikkolaakson metsän hoitamista.
'''

import random     # Moduli satunnaislukujen tuottamiseen
import math    # sin(), cos(), yms.
import pygame    # videopelikirjasto
import katveet   # moduli puiden katveiden pinta-alan laskemiseen
import params    # Viritysparametrejä

# Värit tehdään sekoittamalla punaista, vihreää ja sinistä
# Kutakin väriä voi laittaa 0-255 yksikköä
# Tässä sekoitetaan valoja, ei värejä heijastavia maaleja.
# Vesivärien sekoittamiseen tottuneille tämä oli aluksi tosi outoa.

# Vakiona käytettäviä globaaleja muuttujia.
# Näimä eivät ole niin 'paheksuttavia' kuin alempana esiteltävät globaalit
# muuttujat.

SIN = (0, 0, 255)    # vain sinistä
PUN = (255, 0, 0)
KELT = (255, 255, 0)
VALK = (255, 255, 255)    # valkoisessa on kaikenväristä valoa
HARMAA = (125, 125, 125)  # harmaassakin on kaikenväristä valoa, mutta vähemmän
MAASTOVARI = (25, 25, 10)
LEHTIVARI = (60, 255, 0)    # Vihreää paljon ja vähän punaista
LAHOVARI = (125, 125, 50)   # Ruskeaa


# Muutama globaali muuttuja. Hyvin tehdyssä ohjelmassa ei käytetä globaaleja
# muuttujia, koska niitä käyttäessään tekee helposti virheitä.
# Globaalit muuttujat on tapana kirjoittaa isolla, että ne erottuvat
# koodin seasta.

# AIKA kertoo, paljonko aikaa on kulunut simuloinnin alusta lukien. Tätä tietoa
# tarvitaan monessa eri paikassa koodia. Ohjelman ensimmäistä versiota tehdessä
# oli helppo välittää tieto ajan kulusta globaalin muuttujan avulla.
# Ei tullut korjattua tyylikkäämmäksi.

AIKA = 0     # Simuloitu aika
LOPPU = False    # Simulointi päättyy, kun LOPPU = True
AUTOMAATTI = False    # Automaattinen harvennus


def overlap(a, b):
    '''
    Lasketaan, paljonko puu b varjostaa puuta a.
    '''
    if a.r > 0 and b.r > 0:
        # Lasketaan puiden runkojen välinen etäisyys
        d = math.sqrt((b.y - a.y)**2 + (b.x - a.x)**2)
        # varjostavat, jos lehvästön säteiden summa > puiden välinen etäisyys
        if a.r + b.r > d:
            # Limittäin olevan alueen pinta-ala
            katve = 0.5*katveet.varjo(a.r, b.r, d)
            # Elävä puu varjostaa enemmän kuin pystyynkuollut
            if b.elinvuosia > 0:
                # Lisätään varjostusvaikutukseen, jonka muut lähipuut ovat
                # aiheuttaneet
                # Iso varjostaa pientä pahemmin kuin pieni isoa.
                # yhden naapurin varjo ei tuota täyttä pimentoa,
                # mutta varjostavia naapureita voikin olla kahdeksan
                a.katve += katve*b.r/(a.r+b.r)
            else:
                # pystyynkuollut naapuri varjostaa vain 1/2 siitä, mitä kasvava
                # Luvut perustuvat peikkojen perimätietoon
                a.katve += 0.5*katve*b.r/(a.r+b.r)


def ramp(p1, p2, x):
    '''
    Ramppifuntio. Koodi kertoo kaiken :-)
    Kuvan piirtäminen voi auttaa hahmottamaan
    '''
    (x0, y0) = p1
    (x1, y1) = p2
    if x <= x0:
        return y0
    if x >= x1:
        return y1
    return (x - x0)/(x1 - x0)*(y1 - y0) + y0


def v_pallo(r):
    '''
    r-säteisen pallon tilavuus
    '''
    return 4.0*math.pi*r**3/3.0


class ClDynsim:
    '''
    Integrointia
    '''

    def __init__(self):
        self.maxerr = 0  # Tarkistuksen antama virhe
        self.r_err = 0  # Sattuiko virhe pienelle puulle vai oikean kokoiselle
        self.dt = params.dt

    def integ(self, a, x0):
        '''
        Eulerin algoritmi numeeriseen integrointiin
        if chkdt = True, kokeillaan myös puolta pienemmällä askeleella
        integrointiaskeleen arvioimiseksi.
        Puoliaskeleella edetessä ei lasketa katveita uudelleen, joten tarkistus
        ei ole aivan oikea.

        Tämä on hyvin alkeellinen algoritmi. Esim. kuuraketti-pelissäni
        on yleiskäyttöisempi integrointirutiini.
        '''
        dt2 = self.dt/2
        dx = a.dm(x0)
        xf = x0 + dx*self.dt
        # Integrointiaskeleen pituuden arviointi.
        if params.chkdt:
            zh = x0 + dx*dt2
            dzh = a.dm(zh)
            zf = zh + dzh*dt2
            if zf > 0.0:   # and xf > 0.0
                err = abs((xf-zf)/zf)
            else:
                err = 0.0
            if self.maxerr < err:
                self.maxerr = err
                self.r_err = a.r
            if err > 1.0:
                print('err: {:4.2f}; r: {:4.2f}; m: {:4.2f}; dx: {:4.2f};\
                      dzh: {:4.2f}; katve: {}'.
                      format(err, a.r, a.massa, dx*self.dt, dzh*self.dt,
                             a.katve))
        # Tässä ohjelmassa integroidaan vain puiden kokoa, joten rajoitetaan
        # lopputulos >= 0. Toki dm, puun massan aikaderivaatta, menee nollaksi,
        # kun puu on lahonnut loppuun, mutta integroinnin virheen takia massa saattaa
        # lipsahtaa pikkuisen miinukselle, mikä tuotti ongelmia siellä sun täällä,
        # joten virhe oli helpointa "fiksata" tässä.
        return max(0.0, xf)

    def report(self):
        '''
        Tulostetaan ohjelman päätteeksi konsolille tilastotietoa simuloinnista
        '''
        print('aika: {}'.format(AIKA))
        print('maksimi hiilivarasto: {:4e}; \n keskiarvo hiili {:4e};  polttopuu kaikkiaan {:4e}'.
              format(Metsa.max_varasto, Metsa.ka_hiili, Metsa.tot_polttopuu))
        print('maxerr: {:4e}, vastaava r: {:4e}'.format(self.maxerr, self.r_err))


class ClGraph:
    '''
    Metsän puiden ja tietopylväiden näytölle piirtämiseen tarvittavat funktiot.

    Aluksi meillä oli piirtämiseen tarvittavia pygame - kutsuja eri puolilla
    koodia ja wx_naytto, wy_naytto ja kaikki muut tämän luokan sisäiset muuttujat
    globaaleina muuttujina. Kokoamalla ne tähän olioon, saimme muusta osasta koodia
    vähän selkeämpää.

    Ohjelmakirjastojen käytön aloittaminen on sikamaisen hankalaa.
    Et voi mitenkään päätellä, miten minkäkin python-modulin metodit toimivat,
    mitä niistä tarvitsee tai mitä niitä edes on.
    Monet asiat on manuaaleissa selitetty tavalla, jota ei ymmärrä ennen kuin
    on kokeillut muutamaa esimerkkiä. Kannattaakin aloittaa googlailemalla ja
    kokeilemalla esimerkkejä.
    '''
    def __init__(self):
        pygame.display.init()    # joku init yleensä tarvitaan
        D_Info = pygame.display.Info()    # Tietoa käyttäjän näytöstä
        self.wy_naytto = int(1.0*D_Info.current_h)
        self.wx_naytto = int(1.0*D_Info.current_w)
        self.max_hiili = 0.0
        self.font_metsa = pygame.font.SysFont('robotoslab', 22)
        self.TekstiLaatikot = []
        self.screen = pygame.display.set_mode((self.wx_naytto, self.wy_naytto))
        pygame.display.set_caption("Peikkometsän hiilitase")

    def tausta(self):
        '''
        '''
        self.screen.fill(MAASTOVARI)    # pyyhitään peli-ikkuna tyhjäksi

    def ympyra(self, vari, x, y, r):
        '''
        '''
        pygame.draw.circle(self.screen, vari, (int(x), int(y)), int(r), 1)

    def kiekko(self, vari, x, y, r):
        '''
        '''
        pygame.draw.circle(self.screen, vari, (int(x), int(y)), int(r), 0)
        pygame.draw.circle(self.screen, VALK, (int(x), int(y)), int(r), 1)

    def piirra_varastot(self, hiilivarasto, polttopuu):
        '''
        Piirretään hiilivarastoa kuvaava pylväs ja sen jatkoksi
        varastoidun polttopuun määrä kuvaava pylväs
        '''
          # Metsän puihin varastoitunut hiili
        h1 = int(hiilivarasto/self.max_hiili*self.wy_naytto)
        rect1 = (self.wx_naytto - 75, int(0.95*self.wy_naytto) - h1, 50, h1)
        pygame.draw.rect(self.screen, HARMAA, rect1, 0)
          # Liitereihin varastoitu hiili
        h2 = int(polttopuu/self.max_hiili*self.wy_naytto)
        rect2 = (self.wx_naytto - 75, int(0.95*self.wy_naytto) - (h1+h2), 50, h2)
        if polttopuu < params.puunKulutus:
            vari = PUN
        else:
            vari = KELT
        pygame.draw.rect(self.screen, vari, rect2, 0)

    def lampotolppa(self, lampo):
        '''
        Piirretään kraatterin lämpötilaa kuvaava pylväs
        '''
        global LOPPU
        h = int(lampo/1.0e+10*self.wy_naytto)
        rect = (self.wx_naytto - 150, int(0.5*self.wy_naytto) - h, 50, h)
        if h > 0:
            color = PUN    # Lämpötila yli tavoitteen
        else:
            color = SIN    # Lämpötilan alle tavoitteen
        pygame.draw.rect(self.screen, color, rect, 0)
          # Pysäytetään simulointi, kun tolppa nousee tappiin
        if h > 0.5*self.wy_naytto or h < -0.5*self.wy_naytto:
            LOPPU = True

    def tekstit(self, ohjeet):
        self.TekstiLaatikot = []
        for ohje in ohjeet:
            (teksti, (px, py)) = ohje
            rend_ohje = self.font_metsa.render(teksti, True, (255, 255, 255), (0, 0, 0))
            ix = int(px * self.wx_naytto)
            iy = int(py * self.wy_naytto)
            self.TekstiLaatikot.append((rend_ohje, (ix, iy)))

    def blitTekstit(self):
        '''
        '''
        for laatikko in self.TekstiLaatikot:
            (teksti, paikka) = laatikko
            self.screen.blit(teksti, paikka)

    def blitvuosi(self):
        '''
        Kirjoitetaan vuosiluku näytölle
        '''
        rend_vuosi = self.font_metsa.render(
            "vuosi: {}".format(int(AIKA)), True, (255, 255, 255))
        ixvuosi = int(0.8*self.wx_naytto)
        iyvuosi = int(0.05*self.wy_naytto)
        self.screen.blit(rend_vuosi, (ixvuosi, iyvuosi))

    def render_report(self):
        self.tekstit([
            (' Maksimi hiilivarasto: {:4e} '.format(Metsa.max_varasto), (0.4, 0.45)),
            (' Keskiarvo hiilivarasto {:4e} '.format(Metsa.ka_hiili), (0.4, 0.5)),
            (' Polttopuu kaikkiaan {:4e} '.format(Metsa.tot_polttopuu), (0.4, 0.55))])


class ClPuu():
    '''
    Olioluokan puu kuvaus
    Puut istutetaan ruudukkoon Metsa[i,j] matkan etaisyys välein
    '''
    def __init__(self, i, j, etaisyys):
        '''
        Alkuarvot, jotka puulle annetaan simuloinnin alussa.
        Puun on alussa nukkuva siemen, joka vasta myöhemmin herää kasvamaan.
        Puulla ei ole alussa elinvuosia, ei kokoa ei massaa.
        '''
        self.i = i
        self.j = j
        self.x = self.i*etaisyys + 2.1*etaisyys   # Puun x-koordinaatti
        self.y = self.j*etaisyys + 2.1*etaisyys    # Puun y-koordinaatti
        self.elinvuosia = 0.0    # Jäljellä olevat elinvuodet
        self.katve = 0.0     # Pinta-ala, jonka verran puu on muiden varjossa
        self.vari = LEHTIVARI   # Lehvästön väri
        self.tiheys = 0.1     # Kiinteän puumassan suhde lehvästön kokoon
        self.r = 0.0   # Lehvästön säde
        self.massa = self.tiheys*v_pallo(self.r)
        self.cKasvu = random.uniform(0.025, 0.075)   # Kasvuvauhti
        self.rtaimi = 1.0
        self.leimattu = False    # Onko puu leimattu kaadettavaksi

    def piirra(self):
        '''
        Piirretään kiekko markkeeraamaan puun lehvästöä
        '''
        if self.r >= 2:
              # Ihan pieniä ei kannata piirtää
            if Metsa.iso_pieni:
                  # Piirretään erikseen kiekko kuvaamaan itse puuta ja ympyrä kuvaamaan varjoa
                r = int(self.r/2.0)
                piirrokset.kiekko(self.vari, self.x, self.y, r)
                  # Piiretään kiekon ympärille ympyrä kuvaamaan aluetta,
                  # jota puu varjostaa (ainakin osan päivästä)
                piirrokset.ympyra(self.vari, self.x, self.y, self.r)
            else:
                  # Yhdistetään ylläesitetyt yhdeksi isoksi kiekoksi
                piirrokset.kiekko(self.vari, self.x, self.y, self.r)


    def fkasvu(self, r):
        '''
        Muokataan puun kasvukerrointa puun iän ja koon mukaan.
        Puun kasvu hidastuu eliniän lopulla. Lahoaminen on aluksi
        hidasta, mutta kiihtyy vähitellen.

        Satumetsän puut eivät kasva tiettyä rajaa suuremmiksi ja niiden kasvu
        alkaa hidastua jo ennen ylärajan saavuttamista. Simulaattorissa
        laitoimme puut kasvamaan niin, että niiden välimatka on puolet
        maksimikoosta.

        Parametrit perustuvat peikkovanhuksilta saatuun perimätietoon.
        '''
        c1 = ramp((-30.0, -2.0*self.cKasvu), (30.0, self.cKasvu),
                  self.elinvuosia)
        if self.elinvuosia > 0:
            c2 = ramp((1.5*Metsa.etaisyys, c1), (2.0*Metsa.etaisyys, 0.0), r)
        else:
            c2 = c1
        return c2

    def sade(self):
        '''
        Lehvästön säde massan ja tilavuuden funktiona
        '''
        return math.pow(3.0*self.massa/4.0/self.tiheys, 1/3)

    def varimuutos(self):
        '''
        Mitä pahemmin puu on varjossa, sitä tummempi ja sinertävämpi se on.
        Lahopuu on ruskea.
        Kaadettavaksi leimattu puu on punainen
        '''
        a_valo = max(0.0, self.valossa())
        if a_valo <= 0.0:
              # Täysvarjo lyhentää puun elinikää
            self.elinvuosia -= 0.25*params.dt
          # määritetään lehvästön väri
        if self.leimattu:
            self.vari = PUN
        else:
            if self.elinvuosia > 0.0:
                  # Muutetaan väriä sen mukaan, pahastiko puu on valossa
                (red, green, blue) = self.vari
                green = 10 + int(a_valo*245)
                red = int(a_valo*60)
                self.vari = (red, green, blue)
            else:
                  # Lahopuu on ruskea
                self.vari = LAHOVARI

    def kasvu(self):
        '''
        Puu kasvaa tai lahoaa.
        Lasketaan säteilyä vastaanottava pinta-ala ja muutetaan puun väriä
        sen mukaan.

        Tyhjälle paikalla puhkeaa taimi tietyllä todennäköisyydellä
        '''
          # Jos massa > 0, puu on olemassa, joko kasvaa tai lahoaa
        if self.massa > 0.0:
            self.massa = simu.integ(self, self.massa)
            self.r = self.sade()
            self.elinvuosia -= params.dt
            self.varimuutos()
        else:
              # Massa = 0. Puu hakattu. Uusi taimi nousee 1% todennäköisyydellä
            self.leimattu = False
            if random.randint(0, 100) > 99:
                  # Uusi kasvu alkaa
                self.elinvuosia = random.randint(200, 500)
                self.katve = 0.0
                self.vari = LEHTIVARI
                self.r = self.rtaimi
                self.massa = v_pallo(self.r)*self.tiheys

    def dm(self, m):
        '''
        Puun kasvuvauhti = Puun massan muutoksen aikaderivaatta.
        '''
        if m > 0:
              # Kasvaa tai lahoaa
            r = self.sade()
        else:
              # Kuollut, hakattu
            r = 0.0
          # Auringon säteitä imevä pinta-ala
        A0 = math.pi*r**2
        if self.elinvuosia > 0:
            a_valo = max(0.0, A0 - self.katve)
        else:
              # Kuollut puu. Varjossa olo ei hidasta lahoamista
            a_valo = A0
        return self.fkasvu(r)*a_valo

    def valossa(self):
        '''
        Apufuntio. Kertoo, suuriko suhteellinen osuus lehvästöstä on valossa.
        Tätä tarvitaan mm. lehvästön väriä määritettäessä.
        '''
        if self.r > 0:
            A0 = math.pi*self.r**2
            A = max(0.0, A0 - self.katve)
            return A/A0
        return 0.0


class ClMetsa:
    '''
    Olion metsä kuvaus.

    Huom: Kaikki suuret esitetään peikkoyksiköissä

    puut: Taulukko metsän puista. (Oikeastaan lista listoista)

    max_varasto: Suurin simulaation aikana varastoituneen hiilen määrä

    lämpö: Kraatterin ilmaan varastoitunut lämpö, laakson ilman lämpötilaan
    verrannollinen suure.

    ref: Kraatterin ilman lämpötila pysyy vakiona, kun hiiltä on varastoituneena
    liitereihin ja metsän puihin tätä vastaava määrä.
    Mitä pienempi tämä on, sitä helpompaa on torjua kraatterin ilmaston lämpenemistä.

    poltto: Vauhti, jolla poltamme puuta, jos meillä sitä on liitereissämme.

    ka, N: Hiilivaraston keskiarvo ja sen rekursiivisessa laskemisessa käytettävä
    simulointikierrosten määrä.
    Jos antaa metsän kasvaa ilman hakkuita riittävän pitkään, muodostuvaa keskiarvoa
    voi käyttää pohjana ref-muuttujan määrittämiseen.

    polttopuu, hiilivarasto: Liitereihin ja metsän puihin varastoitu hiilen
    määrä.

    max_varasto: Varastoituneen hiilen enimmäismäärä simuloinnin aikana.

    Varastoitunut hiili tarkoittaa oikeastaan varastoituneen puun massaa.
    Turha konstailla miettimällä, suuriko osa puun massasta on hiiltä.
    '''

    def __init__(self, etaisyys):
        self.etaisyys = etaisyys
        # luodaan lista luokan ClPuu olioita
        self.puut = [[ClPuu(i, j, etaisyys) for j in range(params.NyPuu)]
                     for i in range(params.NxPuu)]
        self.max_varasto = 0.0
        self.lampo = 0.0
        self.N = 0
        self.ka_hiili = 0.0
        self.tot_polttopuu = 0.0
        self.polttopuu = 1.0e+6
        self.hiilivarasto = 0.0
        self.katso_alta = False
        self.iso_pieni = False

    def keskihiili(self):
        '''
        Varastoituneen hiilen keskiarvo
        '''
        self.ka_hiili = (self.N*self.ka_hiili + self.hiilivarasto + self.polttopuu)/(self.N+1)
 #       self.tot_polttopuut += self.ka_polttopuut + self.polttopuu)/(self.N+1)
        self.N += 1

    def lammitys(self):
        '''
        Lasketaan, miten kraatterin ilmasto lämpenee tai jäähtyy.
        Piirretään lämpötilaan verrannollinen pylväs.

        Puun kulutus on vakio, kuitenkin niin, että vuodessa poltetaan enintään
        puolet jäljelläolevasta puumäärästä
        '''
          # Poltetaan polttopuuta, jos sitä on
        if self.polttopuu > 0.0:
              # Poltetaan vuodessa enintään 50% varastosta
            poltto = min(params.puunKulutus, 0.5*self.polttopuu*params.dt)
            self.polttopuu -= poltto*params.dt
          # Kraatterin lämpötilan muutos
        self.lampo += (params.refHiili -
                       (self.hiilivarasto + self.polttopuu))*params.dt
        piirrokset.lampotolppa(self.lampo)

    def piirra(self):
        self.piirra_puut()
        piirrokset.piirra_varastot(self.hiilivarasto, self.polttopuu)

    def piirra_puut(self):
        '''
        Piiretään metsän puut.
        Jos katso_alta = False piirretään pienimmät ensin niin,
        että ne jäävät kuvassa isompien valoon.
        F11 -nappulalla voi vaihtaa katso_alta muuttuja arvoa
        '''
        xx = []
        for row in self.puut:
            for puu in row:
                xx.append(puu)
        ss = sorted(xx, key=lambda puu: puu.r, reverse=Metsa.katso_alta)
        for puu in ss:
            puu.piirra()

    def kasvu(self):
        '''
        Puiden kasvu ja kraatterin lämpötilan muutos yhden integrointiaskelen aikana.
        Päivitetään tilastolliset suureet.

        Lasketaan, paljonko puut varjostavat toisiaan.
        Tässä järjestelmässä tämä on ainoa tilamuuttujien - puiden koko - välinen
        keskinäisvaikutus.
        On tärkeää, että kaikki keskinäisvaikutukset tulee huomioita tilamuuttujien
        derivaattoja laskettaessa ennen integrointia.
        Simulointi toimisi väärin, jos laskisi kunkin puun kasvun erikseen ja sen
        jälkeen laskisi sen varjon vaikutuksen seuraavaan.

        Yleensä dynaamisessa simuloinnissa järjestelmän dynamiikka esitetään
        matriisien ja vektorien avulla niin, että laskenta on helppo tehdä tyylikkäästi oikein.
        Tämä nyt vaan jäi tällaiseksi.

        Kraatterin lämpötilan muutos riippuu puiden massasta. Tämä keskinäisvaikutus on
        onneksi yksisuuntainen, joten ei ole hirveästi väliä, laskeeko lämmitysvaikutuksen
        ennen vai jälkeen puiden kasvun laskemista.
        '''
          # Kraatterin ilmaston muutos.
          # Kun se lasketaan ennen puiden kasvun laskemista, se tulee
          # laskettua hetken t hiilivaraston arvolla
          # Jos se laskettaisin puiden kasvattamisen jälkeen, se tulisi
          # laskettua hetken t+dt hiilivaraston arvolla
        self.lammitys()
        self.hiilivarasto = 0.0
          # Lasketaan, paljonko puut varjostavat toisiaan.
        self.katveet()
          # puiden kasvu ja hiilivaraston päivitys
        for row in self.puut:
            for puu in row:
                puu.kasvu()
                self.hiilivarasto += puu.massa
        self.keskihiili()
        self.max_varasto = max(self.max_varasto,
                               self.hiilivarasto + self.polttopuu)


    def katveet(self):
        '''
         Lasketaan, paljon vierekkäiset puut varjostavat toisiaan.
         Metsän keskellä jokaisella puulla on kahdeksan vierustoveria.
         metsän reunoilla vähemmän.

         Overlap päivittää jokaisen olion 'p' attribuutin 'katve'.
         koodi taitaisi olla selkeämmän näköistä, jos overlap palauttaisi
         katveen arvon ja sijoitus arvolle puu.katve tehtäisiin tässä
        '''
        for i in range(params.NxPuu):
            for j in range(params.NyPuu):
                self.puut[i][j].katve = 0.0
        for i in range(params.NxPuu):
            for j in range(params.NyPuu):
                if i > 0:
                    overlap(self.puut[i][j], self.puut[i-1][j])
                if i < params.NxPuu - 1:
                    overlap(self.puut[i][j], self.puut[i+1][j])
                if j > 0:
                    overlap(self.puut[i][j], self.puut[i][j-1])
                if j < params.NyPuu-1:
                    overlap(self.puut[i][j], self.puut[i][j+1])
                if i > 0 and j > 0:
                    overlap(self.puut[i][j], self.puut[i-1][j-1])
                if i > 0 and j < params.NyPuu - 1:
                    overlap(self.puut[i][j], self.puut[i-1][j+1])
                if i < params.NxPuu-1 and j > 0:
                    overlap(self.puut[i][j], self.puut[i+1][j-1])
                if i < params.NxPuu-1 and j < params.NyPuu - 1:
                    overlap(self.puut[i][j], self.puut[i+1][j+1])

    def leimaa_vanhat(self):
        '''
        Leimataan vanhat puut - elinikää jäljellä alle 15 vuotta - kaadettaviksi.
        Puita voi leimata ja hakata joko käyttäjä tai automaattinen algoritmi.
        '''
        for i in range(params.NxPuu):
            for j in range(params.NyPuu):
                if self.puut[i][j].elinvuosia < 15:
                    self.puut[i][j].leimattu = True

    def hakkaa(self):
        '''
        Kaadetaan leimatut puut polttopuiksi
        '''
        for i in range(params.NxPuu):
            for j in range(params.NyPuu):
                if self.puut[i][j].leimattu:
                    hakkuumaara = self.puut[i][j].massa
                    self.polttopuu += hakkuumaara
                    self.tot_polttopuu = self.tot_polttopuu + hakkuumaara
                    self.puut[i][j].r = 0.0
                    self.puut[i][j].massa = 0.0
                    self.puut[i][j].leimattu = False

    def leimaa_isot(self):
        '''
        Leimataan isoimmat, jo lähes täyteen kokoon kasvaneet puut
        '''
        for row in self.puut:
            for puu in row:
                if puu.r > 1.9*self.etaisyys:
                    puu.leimattu = True

# Vaihtoehto: Hakataan 10% puista suuruusjärjestyksessä isoimmasta lähtien
 #               xx = []
 #       for row in self.puut:
 #           for puu in row:
 #               xx.append(puu)
 #       ss = sorted(xx, key=lambda puu: puu.r, reverse=True)
 #       n = int(0.1*params.NxPuu*NyPuu)
 #       for i in range(n):


    def leimaa_varjot(self):
        '''
        Leimakataan pahiten varjossa olevat puut.
        '''
        for row in self.puut:
            for puu in row:
                if puu.valossa() < 0.1:
                    puu.leimattu = True

    def harvennusLeimaus(self, i0):
        '''
        Leimataan puista joka toinen.
        Jos puurivejä on pariton määrä, harvennus menee tyylikkäästi näin yksinkertaisella
        algoritmilla
        '''
        xx = []
        for row in self.puut:
            for puu in row:
                xx.append(puu)
        for i in range(i0, params.NxPuu*params.NyPuu, 2):
            xx[i].leimattu = True

    def leimaa_kaikki(self):
        '''
        '''
        for row in self.puut:
            for puu in row:
                if puu.r > self.etaisyys/10.0:
                    puu.leimattu = True


class ClAutoharvest:
    '''
    Automaattinen harvennusalgoritmi
    '''
    def __init__(self):
        self.Leimattu = 0

    def Hakkuu(self, Leimikko):
        '''
        Suoritetaan hakkuu 20 vuoden välein
        (x % y tarkoittaa x mod y eli jakojäännöstä)

        Kokeilin myös hakkuustrategiaa, jos metsään mennään, kun polttopuut ovat
        vähissä, mutta se tuntui metsänhoidollisesti älyttömältä

        Kommentoituna on vaihtoehtoisia strategioita. Samalla kertaa voi tehdä
        useammanlaisia hakkuita.

        Huomaa, että hakkuiden välillä täytyy päivittää katveet, jos seuraavan
        hakkuun kriteerinä on varjostus

        Leimauksen ja hakkuun välillä odotetaan niin monta vuotta, että simulointiaikaa
        kuluu sekunnin verran. Tämän ajan puut näkyvät punaisina.

        Koska kaikki funktiot on tarkoitus suorittaa aina kun
        näyttöä päivitetään, eli 'Näyttötaajuus' kertaa sekunnissa, näihin funktioihin
        ei voi laittaa sleep() -tms. kutsua kuin aivan poikkeustapauksissa.

        Odottelun voi ohjelmoida laskurilla, jota päivitetään joka kerta, kun näyttöä
        päivitetään.
        '''
          # if Metsa.polttopuu < params.HakkuuRaja:
        if AIKA % 20 == 0 and not self.Leimattu:
              # On aika leimata
            Leimikko.leimaa_vanhat()
              # Leimikko.katveet()
              # Leimikko.leimaa_varjot()
            self.Leimattu = params.Nayttotaajuus
        elif self.Leimattu > 1:
              # Leimaus tehty. Odotellaan hakkuita
            self.Leimattu -= 1
        else:
              # On aika suorittaa hakkuu
            Leimikko.hakkaa()
            self.Leimattu = 0


def tsekkaaNappaimet():
    '''
    '''
    global LOPPU, AUTOMAATTI
    for event in pygame.event.get():
      # Lopetetaan peli sulkemalla ikkuna tai painamalla ESC
        if event.type == pygame.QUIT:
            LOPPU = True
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_F1:
                Metsa.leimaa_vanhat()
            if event.key == pygame.K_F2:
                Metsa.leimaa_isot()
            if event.key == pygame.K_F3:
                Metsa.harvennusLeimaus(0)
            if event.key == pygame.K_F4:
                Metsa.harvennusLeimaus(1)
            if event.key == pygame.K_F5:
                Metsa.leimaa_kaikki()
            if event.key == pygame.K_F6:
                Metsa.leimaa_varjot()
            if event.key == pygame.K_F12:
                Metsa.katso_alta = not Metsa.katso_alta
            if event.key == pygame.K_F11:
                Metsa.iso_pieni = not Metsa.iso_pieni
            if event.key == pygame.K_F10:
                AUTOMAATTI = not AUTOMAATTI
        if event.type == pygame.KEYUP:
            if event.key == pygame.K_F1:
                Metsa.hakkaa()
            if event.key == pygame.K_F2:
                Metsa.hakkaa()
            if event.key == pygame.K_F3:
                Metsa.hakkaa()
            if event.key == pygame.K_F4:
                Metsa.hakkaa()
            if event.key == pygame.K_F5:
                Metsa.hakkaa()
            if event.key == pygame.K_F6:
                Metsa.hakkaa()
            if event.key == pygame.K_ESCAPE:
                LOPPU = True


##====================================================================
#
# Pääohjelma alkaa tästä
#
# ====================================================================

pygame.init()

# print(pygame.font.get_fonts())

# Luodaan olioita
piirrokset = ClGraph()   # piirtely ja tekstit
simu = ClDynsim()    # integrointi
puiden_etaisyys = piirrokset.wy_naytto/(params.NyPuu+5)
Metsa = ClMetsa(puiden_etaisyys)    # Metsä
AutoHarvest = ClAutoharvest()    # Hakkuualgoritmi

piirrokset.max_hiili = 0.9*Metsa.puut[0][0].tiheys* \
    v_pallo(2*Metsa.etaisyys)*(params.NxPuu + params.NyPuu)

# Näytölle tulevat peliohjeet
piirrokset.tekstit([
    (" Hakkuut: F1: vanhat, F2: isot; F3-F4: joka toinen, F5: kaikki, F6: katveista, F10: automaattinen harvennus ",
     (0.05, 0.85)),
    (" pun/sin: ilman lämpötila; harmaa: puihin varastoitunut hiili; kelt: polttopuu ",
     (0.5, 0.9,)),
    (" F11: Katve erikseen tai ei; F12: Katso ylhäältä tai alhaalta ",
     (0.05, 0.9))])

clock = pygame.time.Clock()
random.seed()

LOPPU = False

# Seuraavaa silmukkaa suoritetaan naytevali kertaa sekunnissa
while not LOPPU:
    AIKA += params.dt
    piirrokset.tausta()    # pyyhkäistään "taulu" (screen) tyhjäksi
    Metsa.kasvu()    # Annetaan puiden kasvaa yhden aika-askelen ajan
    Metsa.piirra()    # Piirretään puut ja ilman lämpötilaa ja hiilivarastojen pylväät
    tsekkaaNappaimet()    # Tarkistetaan, mitä komentoja käyttäjä on antanut
    if AUTOMAATTI:
        AutoHarvest.Hakkuu(Metsa)    # automaattinen harvennus, jos sellaisen aika

    piirrokset.blitTekstit()    # Kirjoitetaan ohjetekstit näytölle
    piirrokset.blitvuosi()    # Kirjoitetaan vuosi

    pygame.display.flip()    # Laitetaan taulu näkyviin ja aletaan piirtää toiselle taululle
    clock.tick(params.Nayttotaajuus)    # Odotellaan, kunnes on kulunut 1/Naytevali sekuntia

# simulointi päättyi. "Kirjoitetaan" raportti
piirrokset.render_report()

# Aloitetaan uudestaan näytön päivitys 1/Naytevali kertaa sekunnissa
# Nyt ei kuitenkaan simuloida vaan näytetään raporttia yms. pientä
LOPPU = False
while not LOPPU:
    for evnt in pygame.event.get():
      # Lopetetaan peli sulkemalla ikkuna tai painamalla ESC
        if evnt.type == pygame.QUIT or evnt.key == pygame.K_ESCAPE:
            LOPPU = True
    piirrokset.blitTekstit()
    piirrokset.blitvuosi()
    pygame.display.flip()
    clock.tick(params.Nayttotaajuus)

pygame.quit()    # Suljetaan pelinäyttö

simu.report()    # Tulostetaan raportti konsolille
